/*******************************************************************************
  MPLAB Harmony Application Source File

  Company:
    Microchip Technology Inc.

  File Name:
    app.c

  Summary:
    This file contains the source code for the MPLAB Harmony application.

  Description:
    This file contains the source code for the MPLAB Harmony application.  It
    implements the logic of the application's state machine and it may call
    API routines of other MPLAB Harmony modules in the system, such as drivers,
    system services, and middleware.  However, it does not call any of the
    system interfaces (such as the "Initialize" and "Tasks" functions) of any of
    the modules in the system or make any assumptions about when those functions
    are called.  That is the responsibility of the configuration-specific system
    files.
 *******************************************************************************/

// *****************************************************************************
// *****************************************************************************
// Section: Included Files
// *****************************************************************************
// *****************************************************************************

#include "app.h"
#include <definitions.h>

//#define DEBUG_MODE

void hci_init(void);
void hci(void);

// *****************************************************************************
// *****************************************************************************
// Section: Global Data Definitions
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
/* Application Data

  Summary:
    Holds application data

  Description:
    This structure holds the application's data.

  Remarks:
    This structure should be initialized by the APP_Initialize function.
    
    Application strings and buffers are be defined outside this structure.
 */

APP_DATA appData __attribute__((coherent)) __attribute__((aligned(4)));

/* buffers */
uint8_t buf[64] __attribute__((coherent)) __attribute__((aligned(4)));
uint8_t buf1[64] __attribute__((coherent)) __attribute__((aligned(4)));
uint8_t bufa[64] __attribute__((coherent)) __attribute__((aligned(4)));

/* flags, passes, and hid_state */
uint8_t l2cap_pass = 0;
bool l2cap_flag = false;
uint8_t sdp_pass = 0;
bool sdp_flag = false;
uint8_t hid_pass = 0;
uint8_t hid_state = 0; //0x00 0x11 0x13 0xff

bool acl_suspned_flag = false; //false: waiting ALC packet, true: need not wait

bool ConnectSwitch;

// *****************************************************************************
// *****************************************************************************
// Section: Application Callback Functions
// *****************************************************************************
// *****************************************************************************

USB_HOST_EVENT_RESPONSE APP_USBHostEventHandler
(
        USB_HOST_EVENT event,
        void * eventData,
        uintptr_t context
        ) {
    /* This function is called by the USB Host whenever a USB Host Layer event
     * has occurred. In this example we only handle the device unsupported event
     * */

    switch (event) {
        case USB_HOST_EVENT_DEVICE_UNSUPPORTED:

            /* The attached device is not supported for some reason */
            break;

        default:
            break;

    }

    return (USB_HOST_EVENT_RESPONSE_NONE);
}

void APP_USBHostGENAttachEventListener(USB_HOST_GEN_OBJ genObj, uintptr_t context) {
    /* This function gets called when the GEN device is attached. Update the
     * application data structure to let the application know that this device
     * is attached */

    appData.deviceIsAttached = true;
    appData.genObj = genObj;
}

USB_HOST_GEN_EVENT_RESPONSE APP_USBHostGENEventHandler
(
        USB_HOST_GEN_HANDLE genHandle,
        USB_HOST_GEN_EVENT event,
        void * eventData,
        uintptr_t context
        ) {
    /* This function is called when a GEN Host event has occurred. A pointer to
     * this function is registered after opening the device. See the call to
     * USB_HOST_GEN_EventHandlerSet() function. */

    USB_HOST_GEN_EVENT_ACM_SET_LINE_CODING_COMPLETE_DATA * setLineCodingEventData;
    USB_HOST_GEN_EVENT_SERIAL_STATE_NOTIFICATION_RECEIVED_DATA *notificationReceiveEventData;
    //   USB_HOST_GEN_EVENT_ACM_SET_CONTROL_LINE_STATE_COMPLETE_DATA * setControlLineStateEventData;
    USB_HOST_GEN_EVENT_WRITE_COMPLETE_DATA * writeCompleteEventData;
    USB_HOST_GEN_EVENT_READ_COMPLETE_DATA * readCompleteEventData;


    switch (event) {
        case USB_HOST_GEN_EVENT_ACM_SET_LINE_CODING_COMPLETE://Class Request

            /* This means the application requested Set Line Coding request is
             * complete. */
            setLineCodingEventData = (USB_HOST_GEN_EVENT_ACM_SET_LINE_CODING_COMPLETE_DATA *) (eventData);

            if (setLineCodingEventData->result == USB_HOST_GEN_RESULT_SUCCESS) {
                appData.HciState = APP_HCI_CLASS_REQUEST_COMPLETE;
            } else {
                appData.state = APP_STATE_ERROR;
            }
            break;

        case USB_HOST_GEN_EVENT_SERIAL_STATE_NOTIFICATION_RECEIVED://EP1 IN

            /* This means the application requested Serial State Notification Get request is
             * complete. */

            notificationReceiveEventData = (USB_HOST_GEN_EVENT_SERIAL_STATE_NOTIFICATION_RECEIVED_DATA *) (eventData);

            if (notificationReceiveEventData->result == USB_HOST_GEN_RESULT_SUCCESS) {
                appData.HciState = APP_HCI_READ_EP1_COMPLETE;
            } else {
                appData.state = APP_STATE_ERROR;
            }
            break;

        case USB_HOST_GEN_EVENT_WRITE_COMPLETE://EP2 OUT

            /* This means an application requested write has completed */
            //           appData.writeTransferDone = true;
            writeCompleteEventData = (USB_HOST_GEN_EVENT_WRITE_COMPLETE_DATA *) (eventData);

            if (writeCompleteEventData->result == USB_HOST_GEN_RESULT_SUCCESS) {
                appData.AclState = APP_ACL_WRITE_EP2_COMPLETE;
            } else {
                appData.state = APP_STATE_ERROR;
            }
            break;

        case USB_HOST_GEN_EVENT_READ_COMPLETE://EP2 IN

            /* This means an application requested write has completed */
            //            appData.readTransferDone = true;
            readCompleteEventData = (USB_HOST_GEN_EVENT_READ_COMPLETE_DATA *) (eventData);
            if (readCompleteEventData->result == USB_HOST_GEN_RESULT_SUCCESS) {
                appData.AclState = APP_ACL_READ_EP2_COMPLETE;
            } else {
                appData.state = APP_STATE_ERROR;
            }
            break;

        case USB_HOST_GEN_EVENT_DEVICE_DETACHED:

            /* The device was detached */
            appData.deviceWasDetached = true;
            break;

        default:
            break;
    }

    return (USB_HOST_GEN_EVENT_RESPONE_NONE);
}

// *****************************************************************************
// *****************************************************************************
// Section: Application Local Functions
// *****************************************************************************
// *****************************************************************************


/* TODO:  Add any necessary local functions.
*/


// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    void APP_Initialize ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Initialize(void) {
    /* Initialize the application state machine */

    appData.state = APP_STATE_BUS_ENABLE;
    appData.HciState = APP_HCI_WAIT;
    appData.AclState = APP_ACL_WAIT;
    //    appData.genHostLineCoding.dwDTERate = APP_HOST_GEN_BAUDRATE_SUPPORTED;
    //    appData.genHostLineCoding.bDataBits = (uint8_t) APP_HOST_GEN_NO_OF_DATA_BITS;
    //    appData.genHostLineCoding.bParityType = (uint8_t) APP_HOST_GEN_PARITY_TYPE;
    //    appData.genHostLineCoding.bCharFormat = (uint8_t) APP_HOST_GEN_STOP_BITS;
    //    appData.controlLineState.dtr = 0;
    //    appData.controlLineState.carrier = 0;
    appData.deviceIsAttached = false;
    appData.deviceWasDetached = false;
    //    appData.readTransferDone = false;
    //    appData.writeTransferDone = false;
    //    appData.controlRequestDone = false;
    //    appData.interruptRequestDone = false;

    /* LED off*/
    LED_Off();

    /* Read RB4 ; if RB4 is Low, erase one pase in flash memory */
    if (CONNECTION_SW_Get() == 0) {
        ConnectSwitch = 0;
        NVM_PageErase(PROGRAM_FLASH_BASE_ADDRESS);
        while(NVM_IsBusy() == true);
    } else ConnectSwitch = 1;

    UART3_WriteString("\n\r ************** YTS Bluetooth Stack **************\n\r\n\r");

}


/******************************************************************************
  Function:
    void APP_Tasks (void )
    void APP_HCI(void)
    void APP_ACL(void)

  Remarks:
    See prototype in app.h.
 */

void APP_Tasks(void) {
    /* Check the application's current state. */
//    USB_HOST_GEN_RESULT result;

    if (appData.deviceWasDetached) {
        /* This means the device is not attached. Reset the application state */

        appData.state = APP_STATE_WAIT_FOR_DEVICE_ATTACH;
        //        appData.readTransferDone = false;
        //        appData.writeTransferDone = false;
        //        appData.controlRequestDone = false;
        //        appData.interruptRequestDone = false;
        appData.deviceWasDetached = false;
    }

    switch (appData.state) {
        case APP_STATE_BUS_ENABLE:

            /* In this state the application enables the USB Host Bus. Note
             * how the GEN Attach event handler are registered before the bus
             * is enabled. */

            USB_HOST_EventHandlerSet(APP_USBHostEventHandler, (uintptr_t) 0);
            USB_HOST_GEN_AttachEventHandlerSet(APP_USBHostGENAttachEventListener, (uintptr_t) 0);
            USB_HOST_BusEnable(0);
            appData.state = APP_STATE_WAIT_FOR_BUS_ENABLE_COMPLETE;
            break;

        case APP_STATE_WAIT_FOR_BUS_ENABLE_COMPLETE:

            /* In this state we wait for the Bus enable to complete */
            if (USB_HOST_BusIsEnabled(0)) {
                appData.state = APP_STATE_WAIT_FOR_DEVICE_ATTACH;
            }
            break;

        case APP_STATE_WAIT_FOR_DEVICE_ATTACH:

            /* In this state the application is waiting for the device to be
             * attached */
            if (appData.deviceIsAttached) {
                /* A device is attached. We can open this device */
                appData.state = APP_STATE_OPEN_DEVICE;
                appData.deviceIsAttached = false;
            }
            break;

        case APP_STATE_OPEN_DEVICE:

            /* In this state the application opens the attached device */
            appData.genHostHandle = USB_HOST_GEN_Open(appData.genObj);
            if (appData.genHostHandle != USB_HOST_GEN_HANDLE_INVALID) {
                /* The driver was opened successfully. Set the event handler
                 * and then go to the next state. */
                USB_HOST_GEN_EventHandlerSet(appData.genHostHandle, APP_USBHostGENEventHandler, (uintptr_t) 0);
                appData.state = APP_STATE_SET_LINE_CODING;
                appData.HciState = APP_HCI_CMD_RESET;
            }
            break;

        case APP_STATE_SET_LINE_CODING:
        case APP_STATE_ERROR:
        default:
            break;
    }
}

void APP_HCI(void) {
    
#ifdef DEBUG_MODE
    uint32_t data_num;
#endif
    
    switch (appData.HciState) {
        case APP_HCI_CMD_RESET:
            hci_init();
            appData.HciState = APP_HCI_ARMED;
            break;

        case APP_HCI_ARMED:
            /* EP1 armed: USB_HOST_GENERAL_READ_EP1 */
            USB_HOST_GEN_SerialStateNotificationGet(appData.genHostHandle, NULL, buf1);
            /* EP2 armed: USB_HOST_GENERAL_Read */
            //            USB_HOST_GEN_Read(appData.genHostHandle, NULL, buf, 64);

            appData.HciState = APP_HCI_WAIT;

            break;

        case APP_HCI_READ_EP1_COMPLETE:

#ifdef DEBUG_MODE
            UART3_WriteString(" HCI: ");
            for (data_num = 0; data_num < buf1[1] + 2; data_num++) {
                PutHex(buf1[data_num]);
                UART3_WriteByte(' ');
            }
            UART3_WriteString("\r\n");
#endif

            hci();

            /* EP1 re-armed: USB_HOST_GENERAL_READ_EP1 */
            USB_HOST_GEN_SerialStateNotificationGet(appData.genHostHandle, NULL, buf1);
            appData.HciState = APP_HCI_WAIT;
            break;

        case APP_HCI_CLASS_REQUEST_COMPLETE:
            break;

        case APP_HCI_WAIT:
            break;

        default:
            break;
    }

}

void APP_ACL(void) {

#ifdef DEBUG_MODE
    uint32_t data_num;
#endif
    
    uint32_t i, length1=0, length2=0;


    switch (appData.AclState) {

        case APP_ACL_READ_EP2_COMPLETE:

#ifdef DEBUG_MODE
            UART3_WriteString(" ACL: ");
            for (data_num = 0; data_num < buf[2] + 4; data_num++) {
                PutHex(buf[data_num]);
                UART3_WriteByte(' ');
            }
            UART3_WriteString("\r\n");
#endif

            if (buf[2] != buf[4] + 4) {//process for a divided packet
                Read_ACL_Command();
                buf[2] = buf[4] + 4;
                appData.AclState = APP_ACL_WAIT;
                break;
            }

            if (buf[1] == 0x10) {//process for a divided packet
                for (i = 0; i < length2; i++) bufa[length1 + 4 + i] = buf[4 + i];
                bufa[3] = (length1 + length2) / 256;
                bufa[2] = (length1 + length2) % 256;
                for (i = 0; i < length1 + length2 + 4; i++) buf[i] = bufa[i];
            }

            l2cap_sdp_hid();

            /* EP2 re-armed: USB_HOST_GENERAL_Read */
            //            USB_HOST_GEN_Read(appData.genHostHandle, NULL, buf, 64);

            appData.AclState = APP_ACL_WAIT;
            break;

        case APP_ACL_WRITE_EP2_COMPLETE:
            break;

        case APP_ACL_WAIT:
            break;

        default:
            break;
    }
}

//********************************************************************************
//********************************************************************************
// Write HCI and ACL functions
//********************************************************************************
//********************************************************************************

void Write_HCI_Command(char *message) {
    uint32_t data_size;
    
#ifdef DEBUG_MODE
    uint32_t data_num;
#endif

    data_size = buf1[2] + 0x03;

#ifdef DEBUG_MODE
    UART3_WriteString(message);
    for (data_num = 0; data_num < data_size; data_num++) {
        PutHex(buf1[data_num]);
        UART3_WriteByte(' ');
    }
    UART3_WriteString("\r\n");
#endif

    /* EP0 write: USB_HOST_GENERAL_ClassReques */
    USB_HOST_GEN_ACM_LineCodingSet(appData.genHostHandle, NULL, buf1, data_size);


}

void Write_ACL_Command(char *message) {
    uint32_t data_size;
    
#ifdef DEBUG_MODE
    uint32_t data_num;
#endif

    data_size = buf[2] + 4;

#ifdef DEBUG_MODE
    UART3_WriteString(message);
    for (data_num = 0; data_num < data_size; data_num++) {
        PutHex(buf[data_num]);
        UART3_WriteByte(' ');
    }
    UART3_WriteString("\r\n");
#endif

    /* EP1 write: USB_HOST_GENERAL_Write */
    USB_HOST_GEN_Write(appData.genHostHandle, NULL, buf, data_size);
}

void Read_ACL_Command(void) {
    /* EP2 re-armed: USB_HOST_GENERAL_Read */
    USB_HOST_GEN_Read(appData.genHostHandle, NULL, buf, 64);
}
//********************************************************************************
//********************************************************************************
// Dispatcher for L2CAP, SDP, and HID layers
//********************************************************************************
//********************************************************************************

void l2cap_sdp_hid(void) {
    l2cap_flag = ((buf[6] == 0x01 && buf[7] == 0x00 && buf[8] != 0x07) || l2cap_pass != 0);
    sdp_flag = ((buf[6] >= 0x40 && buf[8] == 0x06) || sdp_pass != 0);

    //GOTO L2CAP layer****************************
    if (l2cap_flag) l2cap();

        //GOTO SDP layer******************************
    else if (sdp_flag) sdp();

        //GOTO pre_hid function **********************
    else if (hid_state == 0x13) pre_hid();

        //GOTO HID layer******************************
    else if (hid_state == 0xff) hid();

    else return;
}
/*******************************************************************************
 End of File
 */
