/*******************************************************************************
  MPLAB Harmony Application Header File

  Company:
    Microchip Technology Inc.

  File Name:
    app.h

  Summary:
    This header file provides prototypes and definitions for the application.

  Description:
    This header file provides function prototypes and data type definitions for
    the application.  Some of these are required by the system (such as the
    "APP_Initialize" and "APP_Tasks" prototypes) and some of them are only used
    internally by the application (such as the "APP_STATES" definition).  Both
    are defined here for convenience.
*******************************************************************************/

#ifndef _APP_H
#define _APP_H

// *****************************************************************************
// *****************************************************************************
// Section: Included Files
// *****************************************************************************
// *****************************************************************************

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdlib.h>
#include "definitions.h"

// DOM-IGNORE-BEGIN
#ifdef __cplusplus  // Provide C++ Compatibility

extern "C" {

#endif
// DOM-IGNORE-END

// *****************************************************************************
// *****************************************************************************
// Section: Type Definitions
// *****************************************************************************
// *****************************************************************************
//#define APP_HOST_GEN_BAUDRATE_SUPPORTED 9600UL
//#define APP_HOST_GEN_PARITY_TYPE        0
//#define APP_HOST_GEN_STOP_BITS          0
//#define APP_HOST_GEN_NO_OF_DATA_BITS    8

#define PROGRAM_FLASH_BASE_ADDRESS  (NVM_FLASH_START_ADDRESS + (NVM_FLASH_SIZE /2 ))

// *****************************************************************************
/* Application states

  Summary:
    Application states enumeration

  Description:
    This enumeration defines the valid application states.  These states
    determine the behavior of the application at various times.
 */

typedef enum {
    /* Application enabled the bus*/
    APP_STATE_BUS_ENABLE,

    /* Application waits for bus to be enabled */
    APP_STATE_WAIT_FOR_BUS_ENABLE_COMPLETE,

    /* Application waits for GEN Device Attach */
    APP_STATE_WAIT_FOR_DEVICE_ATTACH,

    /* GEN Device is Attached */
    APP_STATE_OPEN_DEVICE,

    /* Set the Line Coding */
    APP_STATE_SET_LINE_CODING,

    /* Application waits to get the device line coding */
    //    APP_STATE_WAIT_FOR_GET_LINE_CODING,

    /* Application is in error state */
    APP_STATE_ERROR

} APP_STATES;

// Hci State

typedef enum {
    APP_HCI_WAIT = 0, // Initialize the acl when a device is attached
    APP_HCI_CMD_RESET, // Device is Attached
    APP_HCI_ARMED,
    APP_HCI_READ_EP1_COMPLETE,
    APP_HCI_CLASS_REQUEST_COMPLETE,
    APP_HCI_END,
} HCI_STATE;

// ACL State

typedef enum {
    APP_ACL_WAIT = 0, // Initialize the acl when a device is attached
    APP_ACL_READ_EP2_COMPLETE,
    APP_ACL_WRITE_EP2_COMPLETE,
    ACL_END,
} ACL_STATE;

// *****************************************************************************
/* Application Data

  Summary:
    Holds application data

  Description:
    This structure holds the application's data.

  Remarks:
    Application strings and buffers are be defined outside this structure.
 */

typedef struct
{
    /* The application's current state */
    APP_STATES state;
    HCI_STATE HciState;
    ACL_STATE AclState;

    /* Array to hold read data */
    //    uint8_t inDataArray[64];

    /* GEN Object */
    USB_HOST_GEN_OBJ genObj;

    /* True if a device is attached */
    bool deviceIsAttached;

    /* True if control request is done */
    //    bool controlRequestDone;

    /* Control Request Result */
    //   USB_HOST_GEN_RESULT controlRequestResult;

    /* True if interrupt request is done */
    //    bool interruptRequestDone;

    /* Interrupt Request Result */
    //    USB_HOST_GEN_RESULT interruptRequestResult;

    /* A GEN Line Coding object */
    //    USB_GEN_LINE_CODING genHostLineCoding;

    /* A Control Line State object*/
    //    USB_GEN_CONTROL_LINE_STATE controlLineState;

    /* Handle to the GEN device. */
    USB_HOST_GEN_HANDLE genHostHandle;

    /* True when a write transfer has complete */
    //    bool writeTransferDone;

    /* Write Transfer Result */
    //    USB_HOST_GEN_RESULT writeTransferResult;

    /* True when a read transfer has complete */
    //    bool readTransferDone;

    /* Read Transfer Result */
    //    USB_HOST_GEN_RESULT readTransferResult;





    /* True if device was detached */
    bool deviceWasDetached;

    /* TODO: Define any additional data used by the application. */

} APP_DATA;

// *****************************************************************************
// *****************************************************************************
// Section: Application Callback Routines
// *****************************************************************************
// *****************************************************************************
/* These routines are called by drivers when certain events occur.
 */
USB_HOST_EVENT_RESPONSE APP_USBHostEventHandler
(
        USB_HOST_EVENT event,
        void * eventData,
        uintptr_t context
        );

void APP_USBHostGENAttachEventListener
(
        USB_HOST_GEN_OBJ genObj,
        uintptr_t context
        );

USB_HOST_GEN_EVENT_RESPONSE APP_USBHostGENEventHandler
(
        USB_HOST_GEN_HANDLE genHandle,
        USB_HOST_GEN_EVENT event,
        void * eventData,
        uintptr_t context
        );

// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    void APP_Initialize ( void )

  Summary:
     MPLAB Harmony application initialization routine.

  Description:
    This function initializes the Harmony application.  It places the
    application in its initial state and prepares it to run so that its
    APP_Tasks function can be called.

  Precondition:
    All other system initialization routines should be called before calling
    this routine (in "SYS_Initialize").

  Parameters:
    None.

  Returns:
    None.

  Example:
    <code>
    APP_Initialize();
    </code>

  Remarks:
    This routine must be called from the SYS_Initialize function.
*/

void APP_Initialize ( void );


/*******************************************************************************
  Functions:
 */

void APP_Tasks(void);
void APP_HCI(void);
void APP_ACL(void);

void Read_ACL_Command(void);
void Write_HCI_Command(char *);
void Write_ACL_Command(char *);

void l2cap_sdp_hid(void);
void l2cap(void);
void sdp(void);
void pre_hid(void);
void hid(void);


#endif /* _APP_H */

//DOM-IGNORE-BEGIN
#ifdef __cplusplus
}
#endif
//DOM-IGNORE-END

/*******************************************************************************
 End of File
 */

