/*******************************************************************************
  USB Host GEN Client Driver Implementation

  Company:
    Microchip Technology Inc.

  File Name:
    usb_host_gen.c

  Summary:
    USB Host GEN Client Driver Implementation

  Description:
    This file contains the implementation of the GEN Client Driver API. It
    should be included in the application if the GEN Host Client Driver
    functionality is desired.
*******************************************************************************/

//DOM-IGNORE-BEGIN
/*******************************************************************************
* Copyright (C) 2018 Microchip Technology Inc. and its subsidiaries.
*
* Subject to your compliance with these terms, you may use Microchip software
* and any derivatives exclusively with Microchip products. It is your
* responsibility to comply with third party license terms applicable to your
* use of third party software (including open source software) that may
* accompany Microchip software.
*
* THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
* EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
* WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
* PARTICULAR PURPOSE.
*
* IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT, SPECIAL, PUNITIVE,
* INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE OF ANY KIND
* WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF MICROCHIP HAS
* BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE. TO THE
* FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN
* ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED THE AMOUNT OF FEES, IF ANY,
* THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR THIS SOFTWARE.
 *******************************************************************************/

#include "usb/usb_host_gen.h"
#include "usb/usb_host_client_driver.h"
#include "usb/usb_host.h"
#include "usb/usb_gen.h"
#include "usb/src/usb_host_gen_local.h"

/************************************************
 * GEN Host Client Driver instance objects. One
 * for each GEN device.
 ************************************************/
 USB_HOST_GEN_INSTANCE_OBJ gUSBHostGENObj[USB_HOST_GEN_INSTANCES_NUMBER];

/***********************************************
 * USB Host GEN Attach Listener Objects
 ***********************************************/
 USB_HOST_GEN_ATTACH_LISTENER_OBJ gUSBHostGENAttachListener[USB_HOST_GEN_ATTACH_LISTENERS_NUMBER];

/************************************************
 * GEN Interface to the host layer
 ************************************************/
USB_HOST_CLIENT_DRIVER gUSBHostGENClientDriver =
{
    .initialize = _USB_HOST_GEN_Initialize,
    .deinitialize = _USB_HOST_GEN_Deinitialize,
    .reinitialize = _USB_HOST_GEN_Reinitialize,
    .interfaceAssign = _USB_HOST_GEN_InterfaceAssign,
    .interfaceRelease = _USB_HOST_GEN_InterfaceRelease,
    .interfaceEventHandler = _USB_HOST_GEN_InterfaceEventHandler,
    .interfaceTasks = _USB_HOST_GEN_InterfaceTasks,
    .deviceEventHandler = _USB_HOST_GEN_DeviceEventHandler,
    .deviceAssign = _USB_HOST_GEN_DeviceAssign,
    .deviceRelease = _USB_HOST_GEN_DeviceRelease,
    .deviceTasks = _USB_HOST_GEN_DeviceTasks
};

// *****************************************************************************
// *****************************************************************************
// GEN Host Client Driver Local function
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
/* Function:  
    USB_HOST_GEN_RESULT _USB_HOST_GEN_HostResutlToGENResultMap
    (
        USB_HOST_RESULT hostResult
    )

  Summary: 
    This function will map the USB Host result to GEN Result.

  Description:
    This function will map the USB Host result to GEN Result.

  Remarks:
    This is a local function and should not be called by directly by the
    application.
*/

USB_HOST_GEN_RESULT _USB_HOST_GEN_HostResutlToGENResultMap
(
    USB_HOST_RESULT result
)
{
    USB_HOST_GEN_RESULT genResult;
  
    switch(result)
    {
        case USB_HOST_RESULT_SUCCESS:
            genResult = USB_HOST_GEN_RESULT_SUCCESS;
            break;
        case USB_HOST_RESULT_FAILURE:
            /* Note the fall through here. This is intentional */
        case USB_HOST_RESULT_PARAMETER_INVALID:
        case USB_HOST_RESULT_PIPE_HANDLE_INVALID:
            genResult = USB_HOST_GEN_RESULT_FAILURE;
            break;
        case USB_HOST_RESULT_REQUEST_BUSY:
            genResult = USB_HOST_GEN_RESULT_BUSY;
            break;
        case USB_HOST_RESULT_REQUEST_STALLED:
            genResult = USB_HOST_GEN_RESULT_REQUEST_STALLED;
            break;
        case USB_HOST_RESULT_TRANSFER_ABORTED:
            genResult = USB_HOST_GEN_RESULT_ABORTED;
            break;
        default:
            genResult = USB_HOST_GEN_RESULT_FAILURE;
            break;
    }
  
    return(genResult);
}

// *****************************************************************************
/* Function:
    int _USB_HOST_GEN_DeviceHandleToInstance
    (
        USB_HOST_DEVICE_CLIENT_HANDLE deviceClientHandle
    );

  Summary:
    This function will return the index of the GEN object that own this device.

  Description:
    This function will return the index of the GEN object that own this device.
    If an instance is not found, the function will return -1.

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

int _USB_HOST_GEN_DeviceHandleToInstance
(
    USB_HOST_DEVICE_CLIENT_HANDLE deviceClientHandle
)
{
    int result = -1;
    int iterator;

    for(iterator = 0; iterator < USB_HOST_GEN_INSTANCES_NUMBER; iterator++)
    {
        if(gUSBHostGENObj[iterator].deviceClientHandle == deviceClientHandle)
        {
            result = iterator;
            break;
        }
    }

    return(result);
}

// *****************************************************************************
/* Function:
    int _USB_HOST_GEN_InterfaceHandleToInstance
    (
        USB_HOST_DEVICE_INTERFACE_HANDLE interfaceHandle
    );

  Summary:
    This function will return the index of the GEN object that own this
    interface.

  Description:
    This function will return the index of the GEN object that owns this
    interface.  If an instance is not found, the function will return -1.

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

int _USB_HOST_GEN_InterfaceHandleToInstance
(
    USB_HOST_DEVICE_INTERFACE_HANDLE interfaceHandle
)
{
    int result = -1;
    int iterator;

    for(iterator = 0; iterator < USB_HOST_GEN_INSTANCES_NUMBER; iterator++)
    {
        if((gUSBHostGENObj[iterator].communicationInterfaceHandle == interfaceHandle) ||
            (gUSBHostGENObj[iterator].dataInterfaceHandle == interfaceHandle))

        {
            result = iterator;
            break;
        }
    }

    return(result);
}

// *****************************************************************************
/* Function:
    int _USB_HOST_GEN_DeviceObjHandleToInstance
    (
        USB_HOST_DEVICE_OBJ_HANDLE deviceObjHandle
    );

  Summary:
    This function will return the index of the GEN object that is associated
    with this device.

  Description:
    This function will return the index of the GEN object that is associated
    with this device.

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

int _USB_HOST_GEN_DeviceObjHandleToInstance
(
    USB_HOST_DEVICE_OBJ_HANDLE deviceObjHandle
)
{
    int result = -1;
    int iterator;

    for(iterator = 0; iterator < USB_HOST_GEN_INSTANCES_NUMBER; iterator++)
    {
        if(gUSBHostGENObj[iterator].deviceObjHandle == deviceObjHandle) 
        {
            result = iterator;
            break;
        }
    }

    return(result);
}

// *****************************************************************************
/* Function:
    void _USB_HOST_GEN_Initialize(void * msdInitData)

  Summary:
    This function is called when the Host Layer is initializing.

  Description:
    This function is called when the Host Layer is initializing.

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

void _USB_HOST_GEN_Initialize(void * data)
{
    int iterator;
    USB_HOST_GEN_INSTANCE_OBJ * genInstanceObj;

    for(iterator = 0; iterator < USB_HOST_GEN_INSTANCES_NUMBER; iterator ++)
    {
        /* Set the pipes handles to invalid */
        genInstanceObj = &gUSBHostGENObj[iterator];
        genInstanceObj->deviceClientHandle = USB_HOST_DEVICE_CLIENT_HANDLE_INVALID;
        genInstanceObj->bulkInPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
        genInstanceObj->bulkOutPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
        genInstanceObj->interruptPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
        genInstanceObj->deviceObjHandle = USB_HOST_DEVICE_OBJ_HANDLE_INVALID;
    }
}

// *****************************************************************************
/* Function:
    void _USB_HOST_GEN_Deinitialize(void)

  Summary:
    This function is called when the Host Layer is deinitializing.

  Description:
    This function is called when the Host Layer is deinitializing.

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

void _USB_HOST_GEN_Deinitialize(void)
{
    /* This function is not implemented in this release of the USB Host stack */
}

// *****************************************************************************
/* Function:
    void _USB_HOST_GEN_Reinitialize(void)

  Summary:
    This function is called when the Host Layer is reinitializing.

  Description:
    This function is called when the Host Layer is reinitializing.

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

void _USB_HOST_GEN_Reinitialize(void * msdInitData)
{
    /* This function is not implemented in this release of the driver */
}

// *****************************************************************************
/* Function:
    void _USB_HOST_GEN_DeviceAssign 
    (
        USB_HOST_DEVICE_CLIENT_HANDLE deviceHandle,
        USB_HOST_DEVICE_OBJ_HANDLE deviceObjHandle,
        USB_DEVICE_DESCRIPTOR * deviceDescriptor
    )
 
  Summary: 
    This function is called when the host layer finds device level class
    subclass protocol match.

  Description:
    This function is called when the host layer finds device level class
    subclass protocol match.

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

void _USB_HOST_GEN_DeviceAssign 
(
    USB_HOST_DEVICE_CLIENT_HANDLE deviceHandle,
    USB_HOST_DEVICE_OBJ_HANDLE deviceObjHandle,
    USB_DEVICE_DESCRIPTOR * deviceDescriptor
)
{
    int iterator;
    USB_HOST_GEN_INSTANCE_OBJ * genInstanceObj = NULL;

    /* If this function is being called, this means that a device class subclass
     * protocol match was obtained.  Look for a free instance */

    for(iterator = 0; iterator < USB_HOST_GEN_INSTANCES_NUMBER; iterator ++)
    {
        /* Search for an available GEN instance object */
        if(!gUSBHostGENObj[iterator].inUse)
        {
            /* Allocate the object */
            genInstanceObj = &gUSBHostGENObj[iterator];
            genInstanceObj->inUse = true;
            genInstanceObj->deviceObjHandle = deviceObjHandle;
            genInstanceObj->deviceClientHandle = deviceHandle;
            break;
        }
    }

    if(genInstanceObj == NULL)
    {
        /* This means an instance could not be allocated. Return the device back
         * to the host */
        USB_HOST_DeviceRelease(deviceHandle);
    }
    else
    {
        /* An instance object was allocated. Check if the device has any
         * configurations */

        if(deviceDescriptor->bNumConfigurations > 0)
        {
            /* This means we have configurations. We can try setting the first 
             * configuration. Also open the control pipe to the device. */
            genInstanceObj->state = USB_HOST_GEN_STATE_SET_CONFIGURATION;
            genInstanceObj->controlPipeHandle = USB_HOST_DeviceControlPipeOpen(deviceObjHandle);
        }
        else
        {
            /* There are no configurations! Move to error state */
            genInstanceObj->state = USB_HOST_GEN_STATE_ERROR;
        }
    }
}
 
// *****************************************************************************
/* Function:
    void _USB_HOST_GEN_DeviceRelease 
    (
        USB_HOST_DEVICE_CLIENT_HANDLE deviceHandle
    )
 
  Summary: 
    This function is called when the device is detached. 

  Description:
    This function is called when the device is detached. 

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

void _USB_HOST_GEN_DeviceRelease
(
    USB_HOST_DEVICE_CLIENT_HANDLE deviceHandle
)
{
    int index;
    USB_HOST_GEN_INSTANCE_OBJ * genInstance;

    /* Find the GEN instance object that owns this device */
    index = _USB_HOST_GEN_DeviceHandleToInstance(deviceHandle);

    if(index >= 0)
    {
        genInstance = &gUSBHostGENObj[index];
        genInstance->inUse = false;

        if(genInstance->bulkInPipeHandle != USB_HOST_PIPE_HANDLE_INVALID)
        {
            /* Close the bulk in pipe and invalidate the pipe handle */
            USB_HOST_DevicePipeClose(genInstance->bulkInPipeHandle);
            genInstance->bulkInPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
        }
        
        if(genInstance->bulkOutPipeHandle != USB_HOST_PIPE_HANDLE_INVALID)
        {
            /* Close the bulk Out pipe and invalidate the pipe handle */
            USB_HOST_DevicePipeClose(genInstance->bulkOutPipeHandle);
            genInstance->bulkOutPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
        }

        if(genInstance->interruptPipeHandle != USB_HOST_PIPE_HANDLE_INVALID)
        {
            /* Close the interruptPipeHandle pipe handle and invalidate the pipe
             * handle */
            USB_HOST_DevicePipeClose(genInstance->interruptPipeHandle);
            genInstance->interruptPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
        }

        if(genInstance->eventHandler != NULL)
        {
            /* Let the client know that the device is detached */
            genInstance->eventHandler((USB_HOST_GEN_HANDLE)(genInstance), 
                    USB_HOST_GEN_EVENT_DEVICE_DETACHED,
                    NULL, genInstance->context);
        }

        /* Release the object */
        genInstance->eventHandler = NULL;
        genInstance->deviceObjHandle = USB_HOST_DEVICE_OBJ_HANDLE_INVALID;
        genInstance->deviceClientHandle = USB_HOST_DEVICE_CLIENT_HANDLE_INVALID;
        
    }
}

// *****************************************************************************
/* Function:
    void _USB_HOST_GEN_DeviceTasks 
    (
        USB_HOST_DEVICE_CLIENT_HANDLE deviceHandle
    )
 
  Summary: 
    This function is called when the host layer want the device to update its
    state. 

  Description:
    This function is called when the host layer want the device to update its
    state. 

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

void _USB_HOST_GEN_DeviceTasks
(
    USB_HOST_DEVICE_CLIENT_HANDLE deviceHandle
)
{
    int index, iterator;
    USB_HOST_GEN_INSTANCE_OBJ * genInstance;
    USB_HOST_RESULT result;
    USB_HOST_REQUEST_HANDLE requestHandle;

    /* Get the index of the device instance */
    index = _USB_HOST_GEN_DeviceHandleToInstance(deviceHandle);

    if(index >= 0)
    {
        /* Get a pointer to the instance */
        genInstance = &gUSBHostGENObj[index];
        
        if(genInstance->inUse)
        {
            /* Instance is valid */
            switch(genInstance->state)
            {
                case USB_HOST_GEN_STATE_NOT_READY:

                    /* The instance is not ready. We dont do anything yet */
                    break;

                case USB_HOST_GEN_STATE_SET_CONFIGURATION:

                    /* The instance should set the configuration. We clear the
                     * controlTransferDone flag. This will be set in the device
                     * event handler when the configuration set event is received. */
                    
                    genInstance->hostRequestDone = false;
                    result = USB_HOST_DeviceConfigurationSet(genInstance->deviceClientHandle, &requestHandle, 
                            0, (uintptr_t)(genInstance));
                    
                    if(result == USB_HOST_RESULT_SUCCESS)
                    {
                        /* The result was successful. Change state to wating for
                         * configuration. */
                        genInstance->state = USB_HOST_GEN_STATE_WAIT_FOR_CONFIGURATION_SET;
                    }
                    break;

                case USB_HOST_GEN_STATE_WAIT_FOR_CONFIGURATION_SET:

                    /* Here we are waiting for the configuration to be set */
                    if(genInstance->hostRequestDone == true) 
                    {
                        if(genInstance->hostRequestResult == USB_HOST_RESULT_SUCCESS)
                        {
                            /* The configuration has been set. Now wait for the host
                             * layer to send the communication and the data
                             * interface to the client driver */

                            genInstance->state = USB_HOST_GEN_STATE_WAIT_FOR_INTERFACES;
                        }
                        else
                        {
                            /* If we could not set the configuration, then state
                             * instance state to error */
                            genInstance->state = USB_HOST_GEN_STATE_ERROR;
                        }
                    }
                    break;

                case USB_HOST_GEN_STATE_WAIT_FOR_INTERFACES:

                    /* Here we wait for both the interfaces to get ready */
                    if((genInstance->interruptPipeHandle != USB_HOST_PIPE_HANDLE_INVALID) &&
                            (genInstance->bulkInPipeHandle != USB_HOST_PIPE_HANDLE_INVALID) &&
                            (genInstance->bulkOutPipeHandle != USB_HOST_PIPE_HANDLE_INVALID))
                    {
                        /* Set the state to ready */
                        genInstance->state = USB_HOST_GEN_STATE_READY;
                        
                        /* We know that the client driver is now ready. We can 
                         * let all the listeners know that the device has been
                         * attached */
                        
                        for(iterator = 0; iterator < USB_HOST_GEN_ATTACH_LISTENERS_NUMBER; iterator ++)
                        {
                            if(gUSBHostGENAttachListener[iterator].inUse)
                            {
                                /* Call the attach listener event handler 
                                 * function. */
                                gUSBHostGENAttachListener[iterator].eventHandler((USB_HOST_GEN_OBJ)(genInstance), 
                                        gUSBHostGENAttachListener[iterator].context);
                            }
                        }
                    }
                            
                    break;
                    
                case USB_HOST_GEN_STATE_READY:
                    
                    /* The GEN client driver is ready to be opened */
                    break;

                case USB_HOST_GEN_STATE_ERROR:

                    /* The instance is an error state. We dont do anything here */
                    break;

            }
        }
    }
}

// *****************************************************************************
/* Function:
    void _USB_HOST_GEN_InterfaceAssign 
    (
        USB_HOST_DEVICE_INTERFACE_HANDLE * interfaces,
        USB_HOST_DEVICE_OBJ_HANDLE deviceObjHandle,
        size_t nInterfaces,
        uint8_t * descriptor
    )

  Summary:
    This function is called when the Host Layer attaches this driver to an
    interface.

  Description:
    This function is called when the Host Layer attaches this driver to an
    interface.

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

void _USB_HOST_GEN_InterfaceAssign 
(
    USB_HOST_DEVICE_INTERFACE_HANDLE * interfaces,
    USB_HOST_DEVICE_OBJ_HANDLE deviceObjHandle,
    size_t nInterfaces,
    uint8_t * descriptor
)
{
    int genInstanceIndex, iterator;
    USB_HOST_GEN_INSTANCE_OBJ * genInstance = NULL;
    USB_INTERFACE_DESCRIPTOR * interfaceDescriptor;
    USB_ENDPOINT_DESCRIPTOR * endpointDescriptor;
    USB_HOST_ENDPOINT_DESCRIPTOR_QUERY endpointDescriptorQuery;
    USB_HOST_INTERFACE_DESCRIPTOR_QUERY interfaceDescriptorQuery; 

    /* This function will be called when there is an interface level match.
     * There are two possible cases here. In case of a simple GEN device, the
     * driver would have matched at the device level in which case the device Client
     * handle will not be invalid. Or else this driver matched at the interface
     * level because this is a composite device. */

    /* If the number of interfaces passed to this function is 1, then we know
     * this is a single interface */

    if(nInterfaces == 1)
    {
        /* One one interface was passed. This means there was device level class
         * subclass protocol match. The device must aleady exist in the system */

        genInstanceIndex = _USB_HOST_GEN_DeviceObjHandleToInstance(deviceObjHandle);
        if(genInstanceIndex >= 0)
        {
            /* Found the instance index that owns this object */
            genInstance = &gUSBHostGENObj[genInstanceIndex];

            /* Get the interface descriptor. We know this is an interface
             * descriptor because the number of interfaces is 1 */
            interfaceDescriptor = (USB_INTERFACE_DESCRIPTOR *)descriptor;
            if ((interfaceDescriptor->bInterfaceClass == 0xe0) &&
                    (interfaceDescriptor->bInterfaceSubClass == 0x01) &&
                    (interfaceDescriptor->bInterfaceProtocol == 0x01)&&
                    (interfaceDescriptor->bInterfaceNumber == 0x00)) 
            {

                /* This interface is the communications class interface. Get the
                 * endpoint number */
                genInstance->communicationInterfaceHandle = interfaces[0];
                genInstance->commInterfaceNumber = interfaceDescriptor->bInterfaceNumber;
                USB_HOST_DeviceEndpointQueryContextClear(&endpointDescriptorQuery);
                endpointDescriptorQuery.transferType = USB_TRANSFER_TYPE_INTERRUPT;
                endpointDescriptorQuery.direction = USB_DATA_DIRECTION_DEVICE_TO_HOST;
                endpointDescriptorQuery.flags = (USB_HOST_ENDPOINT_QUERY_FLAG)(USB_HOST_ENDPOINT_QUERY_BY_TRANSFER_TYPE|USB_HOST_ENDPOINT_QUERY_BY_DIRECTION);
                endpointDescriptor = USB_HOST_DeviceEndpointDescriptorQuery(interfaceDescriptor, &endpointDescriptorQuery);

                /* Did we find the endpoint? */
                if(endpointDescriptor != NULL)
                {
                    /* Found the endpoint. Open the pipe. If the pipe was not
                     * opened, the device will never move to a ready state. */
                    genInstance->interruptPipeHandle = USB_HOST_DevicePipeOpen(genInstance->communicationInterfaceHandle, endpointDescriptor->bEndpointAddress);
                }

                /* This is the data interface */
                genInstance->dataInterfaceHandle = interfaces[0];
                genInstance->dataInterfaceNumber = interfaceDescriptor->bInterfaceNumber;
                USB_HOST_DeviceEndpointQueryContextClear(&endpointDescriptorQuery);
                endpointDescriptorQuery.transferType = USB_TRANSFER_TYPE_BULK;
                endpointDescriptorQuery.direction = USB_DATA_DIRECTION_DEVICE_TO_HOST;
                endpointDescriptorQuery.flags = (USB_HOST_ENDPOINT_QUERY_FLAG)(USB_HOST_ENDPOINT_QUERY_BY_TRANSFER_TYPE|USB_HOST_ENDPOINT_QUERY_BY_DIRECTION);
                endpointDescriptor = USB_HOST_DeviceEndpointDescriptorQuery(interfaceDescriptor, &endpointDescriptorQuery);

                /* Did we find the bulk in point */
                if(endpointDescriptor != NULL)
                {
                    /* Yes we did. Open this pipe */
                    genInstance->bulkInPipeHandle = USB_HOST_DevicePipeOpen(genInstance->dataInterfaceHandle, endpointDescriptor->bEndpointAddress);
                }

                /* Bulk in pipe is opened. Now open the bulk out pipe */
                USB_HOST_DeviceEndpointQueryContextClear(&endpointDescriptorQuery);
                endpointDescriptorQuery.transferType = USB_TRANSFER_TYPE_BULK;
                endpointDescriptorQuery.direction = USB_DATA_DIRECTION_HOST_TO_DEVICE;
                endpointDescriptorQuery.flags = (USB_HOST_ENDPOINT_QUERY_FLAG)(USB_HOST_ENDPOINT_QUERY_BY_TRANSFER_TYPE|USB_HOST_ENDPOINT_QUERY_BY_DIRECTION);
                endpointDescriptor = USB_HOST_DeviceEndpointDescriptorQuery(interfaceDescriptor, &endpointDescriptorQuery);

                /* Did we find the pipe */
                if(endpointDescriptor != NULL)
                {
                    /* Yes we did. Open this pipe */
                    genInstance->bulkOutPipeHandle = USB_HOST_DevicePipeOpen(genInstance->dataInterfaceHandle, 
                            endpointDescriptor->bEndpointAddress);
                }
            } else if ((interfaceDescriptor->bInterfaceClass == 0xe0) &&
                    (interfaceDescriptor->bInterfaceSubClass == 0x01) &&
                    (interfaceDescriptor->bInterfaceProtocol == 0x01) &&
                    (interfaceDescriptor->bInterfaceNumber == 0x01)) {
                //return;
            } 
            else
            {
                /* Dont know what this interface is. Return it back */
                USB_HOST_DeviceInterfaceRelease(interfaces[0]);
            }
        }
        else
        {
            /* This is an error case. The instance should exist. We return this
             * interface back to the host. */
            USB_HOST_DeviceInterfaceRelease(interfaces[0]);
        }
    }
    else if(nInterfaces > 1)
    {
        /* Then this means that this is an IAD. We first assign a GEN instance
         * to this device */

        for(iterator = 0; iterator < USB_HOST_GEN_INSTANCES_NUMBER; iterator ++)
        {
            if(!gUSBHostGENObj[iterator].inUse)
            {
                genInstance = &gUSBHostGENObj[iterator];
                genInstance->inUse = true;
                break;
            }
        }

        if(genInstance == NULL)
        {
            /* This means we could not find an instance. Release all the
             * interfaces in this group */

            for(iterator = 0; iterator < nInterfaces; iterator ++)
            {
                USB_HOST_DeviceInterfaceRelease(interfaces[iterator]);
            }
        }
        else
        {
            /* Save the device object handle and open the control pipe */
            genInstance->deviceObjHandle = deviceObjHandle;
            genInstance->controlPipeHandle = USB_HOST_DeviceControlPipeOpen(deviceObjHandle);

            /* An instance is assigned. The descriptor will be a pointer to the
             * IAD. Lets get the first interface descriptor in the IAD group and
             * see which interface this is. */ 

            USB_HOST_DeviceInterfaceQueryContextClear(&interfaceDescriptorQuery);
            interfaceDescriptorQuery.flags = USB_HOST_INTERFACE_QUERY_ANY;

            /* We know that we need two interfaces */
            for(iterator = 0; iterator < 2; iterator ++)
            {
                /* We need to search for two interface descriptors */
                interfaceDescriptor = USB_HOST_DeviceGeneralInterfaceDescriptorQuery
                        ((USB_INTERFACE_ASSOCIATION_DESCRIPTOR *)(descriptor), &interfaceDescriptorQuery);
                
                /* If we have a valid interface descriptor find out its type */
                if(interfaceDescriptor != NULL)
                {
                    if((interfaceDescriptor->bInterfaceClass == USB_GEN_COMMUNICATIONS_INTERFACE_CLASS_CODE) &&
                            (interfaceDescriptor->bInterfaceSubClass == USB_GEN_SUBCLASS_ABSTRACT_CONTROL_MODEL) &&
                            (interfaceDescriptor->bInterfaceProtocol == USB_GEN_PROTOCOL_AT_V250))
                    {
                        /* We found the communication class */
                        genInstance->commInterfaceNumber = interfaceDescriptor->bInterfaceNumber;
                        genInstance->communicationInterfaceHandle = interfaces[iterator];

                        /* Create the endpoint query */
                        USB_HOST_DeviceEndpointQueryContextClear(&endpointDescriptorQuery);
                        endpointDescriptorQuery.transferType = USB_TRANSFER_TYPE_INTERRUPT;
                        endpointDescriptorQuery.direction = USB_DATA_DIRECTION_DEVICE_TO_HOST;
                        endpointDescriptorQuery.flags = (USB_HOST_ENDPOINT_QUERY_FLAG)(USB_HOST_ENDPOINT_QUERY_BY_TRANSFER_TYPE|USB_HOST_ENDPOINT_QUERY_BY_DIRECTION);
                        endpointDescriptor = USB_HOST_DeviceEndpointDescriptorQuery(interfaceDescriptor, &endpointDescriptorQuery);

                        if(endpointDescriptor != NULL)
                        {
                            /* Open the pipe */
                            genInstance->interruptPipeHandle = USB_HOST_DevicePipeOpen(genInstance->communicationInterfaceHandle, 
                                    endpointDescriptor->bEndpointAddress);
                        }
                        else
                        {
                            /* Make sure that the pipe handle stays invalid if
                             * we could not open the pipe */
                            genInstance->interruptPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
                        }
                    }
                    else if ((interfaceDescriptor->bInterfaceClass == USB_GEN_DATA_INTERFACE_CLASS_CODE) && 
                            (interfaceDescriptor->bInterfaceSubClass == 0x00) && 
                            (interfaceDescriptor->bInterfaceProtocol == USB_GEN_PROTOCOL_NO_CLASS_SPECIFIC))
                    {
                        /* We found the data class */

                        genInstance->dataInterfaceHandle = interfaces[iterator];
                        genInstance->dataInterfaceNumber = interfaceDescriptor->bInterfaceNumber;

                        /* Get the bulk in endpoint */ 
                        USB_HOST_DeviceEndpointQueryContextClear(&endpointDescriptorQuery);
                        endpointDescriptorQuery.transferType = USB_TRANSFER_TYPE_BULK;
                        endpointDescriptorQuery.direction = USB_DATA_DIRECTION_DEVICE_TO_HOST;
                        endpointDescriptorQuery.flags = (USB_HOST_ENDPOINT_QUERY_FLAG)(USB_HOST_ENDPOINT_QUERY_BY_TRANSFER_TYPE|USB_HOST_ENDPOINT_QUERY_BY_DIRECTION);
                        endpointDescriptor = USB_HOST_DeviceEndpointDescriptorQuery(interfaceDescriptor, &endpointDescriptorQuery);

                        if(endpointDescriptor != NULL)
                        {
                            genInstance->bulkInPipeHandle = USB_HOST_DevicePipeOpen(genInstance->dataInterfaceHandle, 
                                    endpointDescriptor->bEndpointAddress);
                        }
                        else
                        {
                            /* Make the pipe handle invalid */
                            genInstance->bulkInPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
                        }

                        /* Get the bulk in endpoint */ 
                        USB_HOST_DeviceEndpointQueryContextClear(&endpointDescriptorQuery);
                        endpointDescriptorQuery.transferType = USB_TRANSFER_TYPE_BULK;
                        endpointDescriptorQuery.direction = USB_DATA_DIRECTION_HOST_TO_DEVICE;
                        endpointDescriptorQuery.flags = (USB_HOST_ENDPOINT_QUERY_FLAG)(USB_HOST_ENDPOINT_QUERY_BY_TRANSFER_TYPE|USB_HOST_ENDPOINT_QUERY_BY_DIRECTION);
                        endpointDescriptor = USB_HOST_DeviceEndpointDescriptorQuery(interfaceDescriptor, &endpointDescriptorQuery);

                        if(endpointDescriptor != NULL)
                        {
                            genInstance->bulkOutPipeHandle = USB_HOST_DevicePipeOpen(genInstance->dataInterfaceHandle, 
                                    endpointDescriptor->bEndpointAddress);
                        }
                        else
                        {
                            /* Make the pipe handle invalid */
                            genInstance->bulkOutPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
                        }

                    }
                }
                else
                {
                    /* There have to be at least two interface descriptors */
                }
            }

            /* Now check if we can move the host client driver instance to a
             * ready state */

            if((genInstance->interruptPipeHandle != USB_HOST_PIPE_HANDLE_INVALID) &&
                    (genInstance->bulkInPipeHandle != USB_HOST_PIPE_HANDLE_INVALID) &&
                    (genInstance->bulkOutPipeHandle != USB_HOST_PIPE_HANDLE_INVALID))
            {
                /* All the pipes are opened. The client driver is ready */
                genInstance->state = USB_HOST_GEN_STATE_READY;

                /* We know that the client driver is now ready. We can 
                 * let all the listeners know that the device has been
                 * attached.  */

                for(iterator = 0; iterator < USB_HOST_GEN_ATTACH_LISTENERS_NUMBER; iterator ++)
                {
                    if(gUSBHostGENAttachListener[iterator].inUse)
                    {
                        /* Call the attach listener event handler 
                         * function. */
                        gUSBHostGENAttachListener[iterator].eventHandler((USB_HOST_GEN_OBJ)(genInstance), 
                                gUSBHostGENAttachListener[iterator].context);
                    }
                }
            }
            else
            {
                /* Something went wrong. */
                genInstance->state = USB_HOST_GEN_STATE_ERROR;
            }
        }
    }
}

// *****************************************************************************
/* Function:
    USB_HOST_DEVICE_INTERFACE_EVENT_RESPONSE _USB_HOST_GEN_InterfaceEventHandler
    (
        USB_HOST_DEVICE_INTERFACE_HANDLE interfaceHandle,
        USB_HOST_DEVICE_INTERFACE_EVENT event,
        void * eventData,
        uintptr_t context
    )

  Summary:
    This function is called when the Host Layer generates interface level
    events. 

  Description:
    This function is called when the Host Layer generates interface level
    events. 

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

USB_HOST_DEVICE_INTERFACE_EVENT_RESPONSE _USB_HOST_GEN_InterfaceEventHandler
(
    USB_HOST_DEVICE_INTERFACE_HANDLE interfaceHandle,
    USB_HOST_DEVICE_INTERFACE_EVENT event,
    void * eventData,
    uintptr_t context
)
{
    int genIndex;
    USB_HOST_GEN_INSTANCE_OBJ * genInstance;
    USB_HOST_DEVICE_INTERFACE_EVENT_TRANSFER_COMPLETE_DATA * dataTransferEvent;
    USB_HOST_GEN_EVENT genEvent;
    USB_HOST_GEN_EVENT_WRITE_COMPLETE_DATA genTransferCompleteData;

    /* Find out to which GEN Instance this interface belongs */
    genIndex = _USB_HOST_GEN_InterfaceHandleToInstance(interfaceHandle);
    genInstance = &gUSBHostGENObj[genIndex];
    genEvent = (USB_HOST_GEN_EVENT)(context); 

    switch(event)
    {
        case USB_HOST_DEVICE_INTERFACE_EVENT_TRANSFER_COMPLETE:

            /* This means a data transfer has completed */
            dataTransferEvent = (USB_HOST_DEVICE_INTERFACE_EVENT_TRANSFER_COMPLETE_DATA *)(eventData);
            genTransferCompleteData.transferHandle = dataTransferEvent->transferHandle;
            genTransferCompleteData.result = _USB_HOST_GEN_HostResutlToGENResultMap(dataTransferEvent->result);
            genTransferCompleteData.length = dataTransferEvent->length;

            if(genInstance->eventHandler != NULL)
            {
                genInstance->eventHandler((USB_HOST_GEN_HANDLE)(genInstance), genEvent, 
                        &genTransferCompleteData, genInstance->context);
            }

            break;

        default:
            break;
    }
    
    return(USB_HOST_DEVICE_INTERFACE_EVENT_RESPONSE_NONE);
}

// *****************************************************************************
/* Function:
    void USB_HOST_GEN_InterfaceTasks
    (
        USB_HOST_DEVICE_INTERFACE_HANDLE interfaceHandle
    )

  Summary:
    This function is called by the Host Layer to update the state of this
    driver.

  Description:
    This function is called by the Host Layer to update the state of this
    driver.

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

void _USB_HOST_GEN_InterfaceTasks
(
    USB_HOST_DEVICE_INTERFACE_HANDLE interfaceHandle
)
{
}

// *****************************************************************************
/* Function:
    void USB_HOST_GEN_InterfaceRelease
    (
        USB_HOST_DEVICE_INTERFACE_HANDLE interfaceHandle
    )

  Summary:
    This function is called when the Host Layer detaches this driver from an
    interface.

  Description:
    This function is called when the Host Layer detaches this driver from an
    interface.

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

void _USB_HOST_GEN_InterfaceRelease
(
    USB_HOST_DEVICE_INTERFACE_HANDLE interfaceHandle
)
{
    int genIndex;
    USB_HOST_GEN_INSTANCE_OBJ * genInstance;
    
    /* Get the instance associated with this interface */
    genIndex = _USB_HOST_GEN_InterfaceHandleToInstance(interfaceHandle);
    
    if(genIndex >= 0)
    {
        /* Get the pointer to the instance object */
        genInstance = &gUSBHostGENObj[genIndex];
        
        if(genInstance->inUse)
        {
            genInstance->inUse = false;

            if(genInstance->bulkInPipeHandle != USB_HOST_PIPE_HANDLE_INVALID)
            {
                /* Close the bulk in pipe and invalidate the pipe handle */
                USB_HOST_DevicePipeClose(genInstance->bulkInPipeHandle);
                genInstance->bulkInPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
            }

            if(genInstance->bulkOutPipeHandle != USB_HOST_PIPE_HANDLE_INVALID)
            {
                /* Close the bulk Out pipe and invalidate the pipe handle */
                USB_HOST_DevicePipeClose(genInstance->bulkOutPipeHandle);
                genInstance->bulkOutPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
            }

            if(genInstance->interruptPipeHandle != USB_HOST_PIPE_HANDLE_INVALID)
            {
                /* Close the interruptPipeHandle pipe handle and invalidate the pipe
                 * handle */
                USB_HOST_DevicePipeClose(genInstance->interruptPipeHandle);
                genInstance->interruptPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
            }

            if(genInstance->eventHandler != NULL)
            {
                /* Let the client know that the device is detached */
                genInstance->eventHandler((USB_HOST_GEN_HANDLE)(genInstance), 
                        USB_HOST_GEN_EVENT_DEVICE_DETACHED,
                        NULL, genInstance->context);
            }

            /* Release the object */
            genInstance->eventHandler = NULL;
            genInstance->deviceObjHandle = USB_HOST_DEVICE_OBJ_HANDLE_INVALID;
            genInstance->deviceClientHandle = USB_HOST_DEVICE_CLIENT_HANDLE_INVALID;
        }
    }
}

// *****************************************************************************
/* Function:
    USB_HOST_DEVICE_INTERFACE_EVENT_RESPONSE _USB_HOST_GEN_DeviceEventHandler
    (
        USB_HOST_DEVICE_CLIENT_HANDLE deviceHandle,
        USB_HOST_DEVICE_EVENT event,
        void * eventData,
        uintptr_t context
    )

  Summary:
    This function is called when the Host Layer generates device level
    events. 

  Description:
    This function is called when the Host Layer generates device level
    events. 

  Remarks:
    This is a local function and should not be called directly by the
    application.
*/

USB_HOST_DEVICE_EVENT_RESPONSE _USB_HOST_GEN_DeviceEventHandler
(
    USB_HOST_DEVICE_CLIENT_HANDLE deviceHandle,
    USB_HOST_DEVICE_EVENT event,
    void * eventData,
    uintptr_t context
)
{
    /* The event context is the pointer to the GEN Instance Object */
    USB_HOST_GEN_INSTANCE_OBJ * genInstance = (USB_HOST_GEN_INSTANCE_OBJ *)(context);
    USB_HOST_DEVICE_EVENT_CONFIGURATION_SET_DATA * configSetEventData;

    switch(event)
    {
        case USB_HOST_DEVICE_EVENT_CONFIGURATION_SET:

            /* This means the configuration was set. Update the instance
             * variables to let the main state machine know. */
            configSetEventData = (USB_HOST_DEVICE_EVENT_CONFIGURATION_SET_DATA *)(eventData);
            genInstance->hostRequestResult =  configSetEventData->result;
            genInstance->hostRequestDone = true;
            break;

        case USB_HOST_DEVICE_EVENT_CONFIGURATION_DESCRIPTOR_GET_COMPLETE:
            break;

        default:
            break;
    }
    
    return(USB_HOST_DEVICE_EVENT_RESPONSE_NONE);
}

// *****************************************************************************
// *****************************************************************************
// Public Functions
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
// *****************************************************************************
// GEN Host Client Driver Public function
// *****************************************************************************
// *****************************************************************************

// ****************************************************************************
/* Function:
    USB_HOST_GEN_RESULT USB_HOST_GEN_AttachEventHandlerSet
    (
        USB_HOST_GEN_ATTACH_EVENT_HANDLER eventHandler,
        uintptr_t context
    );
           
  Summary:
    This function will set an attach event handler.

  Description:
    This function will set an attach event handler. The attach event handler
    will be called when a GEN device has been attached. The context will be
    returned in the event handler. This function should be called before the bus
    has been enabled.

  Remarks:
    Refer to usb_host_gen.h for usage information.
*/

USB_HOST_GEN_RESULT USB_HOST_GEN_AttachEventHandlerSet
(
    USB_HOST_GEN_ATTACH_EVENT_HANDLER eventHandler,
    uintptr_t context
)
{
    int iterator;

    USB_HOST_GEN_RESULT result = USB_HOST_GEN_RESULT_FAILURE;
    USB_HOST_GEN_ATTACH_LISTENER_OBJ * attachListener;
    
    if(eventHandler == NULL)
    {
        result = USB_HOST_GEN_RESULT_INVALID_PARAMETER;
    }
    else
    {
        /* Search for free listener object */
        for(iterator = 0; iterator < USB_HOST_GEN_ATTACH_LISTENERS_NUMBER; iterator ++)
        {
            if(!gUSBHostGENAttachListener[iterator].inUse)
            {
                /* Found a free object */
                attachListener = &gUSBHostGENAttachListener[iterator];
                attachListener->inUse = true;
                attachListener->eventHandler = eventHandler;
                attachListener->context = context;
                result = USB_HOST_GEN_RESULT_SUCCESS;
                break;
            }
        }
    }

    return(result);
}

// ****************************************************************************
/* Function:
    USB_HOST_GEN_HANDLE USB_HOST_GEN_Open
    (
        USB_HOST_GEN_OBJ genDeviceObj
    );
           
  Summary:
    This function opens the specified GEN device.

  Description:
    This function will open the specified GEN device. Once opened, the GEN
    device can be accessed via the handle which this function returns. The
    genDeviceObj parameter is the value returned in the
    USB_HOST_GEN_ATTACH_EVENT_HANDLER event handling function.

  Remarks:
    Refer to usb_host_gen.h for usage information.
*/

USB_HOST_GEN_HANDLE USB_HOST_GEN_Open
(
    USB_HOST_GEN_OBJ genDeviceObj
)
{
    USB_HOST_GEN_HANDLE result = USB_HOST_GEN_HANDLE_INVALID;
    USB_HOST_GEN_INSTANCE_OBJ * genInstance;

    /* The present implementation is a single client implementation only */

    if(genDeviceObj != 0)
    {
        genInstance = (USB_HOST_GEN_INSTANCE_OBJ *)genDeviceObj;
        if((genInstance->inUse) && (genInstance->state == USB_HOST_GEN_STATE_READY))
        {
            result = (USB_HOST_GEN_HANDLE)(genDeviceObj);
        }
    }

    return(result);
}

// *****************************************************************************
/* Function:
    USB_HOST_GEN_RESULT USB_HOST_GEN_EventHandlerSet
    (
        USB_HOST_GEN_HANDLE handle,
        USB_HOST_GEN_EVENT_HANDLER eventHandler,
        uintptr_t context
    );

  Summary:
    Registers an event handler with the GEN Host Client Driver.

  Description:
    This function registers a client specific GEN Host Client Driver event
    handler. The GEN Host Client Driver will call this function with relevant
    event and assocaite event data, in response to command requests and data
    transfers that have been scheduled by the client.

  Remarks:
    Refer to usb_host_gen.h for usage information.
*/

USB_HOST_GEN_RESULT USB_HOST_GEN_EventHandlerSet
(
    USB_HOST_GEN_HANDLE handle,
    USB_HOST_GEN_EVENT_HANDLER eventHandler,
    uintptr_t context
)
{
    USB_HOST_GEN_RESULT result = USB_HOST_GEN_RESULT_HANDLE_INVALID;
    USB_HOST_GEN_INSTANCE_OBJ * genInstance = (USB_HOST_GEN_INSTANCE_OBJ *)(handle);

    if(genInstance != NULL)
    {
        genInstance->eventHandler = eventHandler;
        genInstance->context = context;
        result = USB_HOST_GEN_RESULT_SUCCESS;
    }

    return(result);
}

// ****************************************************************************
/* Function:
    void USB_HOST_GEN_Close
    (
        USB_HOST_GEN_HANDLE genDeviceHandle
    );
           
  Summary:
    This function closes the GEN device.

  Description:
    This function will close the open GEN device. This closes the association
    between the application entity that opened the device and device. The driver
    handle becomes invalid.

  Remarks:
    None.
*/

void USB_HOST_GEN_Close
(
    USB_HOST_GEN_HANDLE genDeviceHandle
)
{
    USB_HOST_GEN_INSTANCE_OBJ * genInstance = (USB_HOST_GEN_INSTANCE_OBJ *)(genDeviceHandle);

    if(genInstance != NULL)
    {
        /* In this release of the GEN driver, the close function does not do
         * much. If the client registered an event handler, then this is set to
         * NULL. */

        genInstance->eventHandler = NULL;
    }
}

// ****************************************************************************
/* Function:
    USB_HOST_GEN_RESULT USB_HOST_GEN_Write
    (
        USB_HOST_GEN_HANDLE handle,
        USB_HOST_GEN_TRANSFER_HANDLE * transferHandle,
        void * data,
        size_t size
    );
           
  Summary:
    This function will write data to the attached device.

  Description:
    This function will write data to the attached GEN device. The function will
    write size amount of bytes. If the request was accepted, transferHandle will
    contain a valid transfer handle, else it will contain
    USB_HOST_GEN_TRANSFER_HANDLE_INVALID. The completion of the request will be
    indicated by the USB_HOST_GEN_EVENT_WRITE_COMPLETE event. The transfer
    handle will be returned in the event. 

  Remarks:
    Refer to usb_host_gen.h for usage information.
*/

USB_HOST_GEN_RESULT USB_HOST_GEN_Write
(
    USB_HOST_GEN_HANDLE handle,
    USB_HOST_GEN_TRANSFER_HANDLE * transferHandle,
    void * data,
    size_t size
)
{
    USB_HOST_GEN_INSTANCE_OBJ * genInstance;
    USB_HOST_GEN_TRANSFER_HANDLE * tempTransferHandle, localTransferHandle;
    USB_HOST_GEN_RESULT genResult = USB_HOST_GEN_RESULT_FAILURE;
    USB_HOST_RESULT hostResult;

    genInstance = (USB_HOST_GEN_INSTANCE_OBJ *)handle;

    if(genInstance == NULL)
    {
        /* This handle is not valid */
        genResult = USB_HOST_GEN_RESULT_HANDLE_INVALID;
    }
    else
    {
        /* Check if the specified transfer handle holder is NULL, if so use a local
         * transfer handle holder */

        tempTransferHandle = (transferHandle == NULL) ? &localTransferHandle: transferHandle;

        if(!genInstance->inUse)
        {
            /* This object is not valid */
            genResult = USB_HOST_GEN_RESULT_DEVICE_UNKNOWN;
        }
        else
        {
            if(genInstance->state != USB_HOST_GEN_STATE_READY)
            {
                /* The instance is not ready for requests */
                genResult = USB_HOST_GEN_RESULT_BUSY;
            }
            else
            {
                if((size != 0) && (data == NULL))
                {
                    /* Input paramters are not valid */
                    genResult = USB_HOST_GEN_RESULT_INVALID_PARAMETER;
                }
                else
                {
                    /* The context for the transfer is the event that needs to
                     * be sent to the application. In this case the event to be
                     * sent to the application when the transfer completes is
                     * USB_HOST_GEN_EVENT_WRITE_COMPLETE */
                    
                    hostResult = USB_HOST_DeviceTransfer(genInstance->bulkOutPipeHandle, tempTransferHandle, data, size, (uintptr_t)(USB_HOST_GEN_EVENT_WRITE_COMPLETE));
                    genResult = _USB_HOST_GEN_HostResutlToGENResultMap(hostResult);
                }
            }
        }
    }

    return(genResult);
}

// ****************************************************************************
/* Function:
    USB_HOST_GEN_RESULT USB_HOST_GEN_Read
    (
        USB_HOST_GEN_HANDLE handle,
        USB_HOST_GEN_TRANSFER_HANDLE * transferHandle,
        void * data,
        size_t size
    );
           
  Summary:
    This function will read data from the attached device.

  Description:
    This function will read data from the attached GEN device. The function will
    try to read size amount of bytes but will stop reading when the device
    terminates the USB transfer (sends a short packet or a ZLP). If the request
    was accepted, transferHandle will contain a valid transfer handle, else it
    will contain USB_HOST_GEN_TRANSFER_HANDLE_INVALID. The completion of the
    request will be indicated by the USB_HOST_GEN_EVENT_READ_COMPLETE
    event. The transfer handle will be returned in the event. 

  Remarks:
    None.
*/
USB_HOST_GEN_RESULT USB_HOST_GEN_Read
(
    USB_HOST_GEN_HANDLE handle,
    USB_HOST_GEN_TRANSFER_HANDLE * transferHandle,
    void * data,
    size_t size
)
{
    USB_HOST_GEN_INSTANCE_OBJ * genInstance;
    USB_HOST_GEN_TRANSFER_HANDLE * tempTransferHandle, localTransferHandle;
    USB_HOST_GEN_RESULT genResult = USB_HOST_GEN_RESULT_FAILURE;
    USB_HOST_RESULT hostResult;

    genInstance = (USB_HOST_GEN_INSTANCE_OBJ *)handle;

    if(genInstance == NULL)
    {
        /* This handle is not valid */
        genResult = USB_HOST_GEN_RESULT_HANDLE_INVALID;
    }
    else
    {
        /* Check if the specified transfer handle holder is NULL, if so use a local
         * transfer handle holder */

        tempTransferHandle = (transferHandle == NULL) ? &localTransferHandle: transferHandle;

        if(!genInstance->inUse)
        {
            /* This object is not valid */
            genResult = USB_HOST_GEN_RESULT_DEVICE_UNKNOWN;
        }
        else
        {
            if(genInstance->state != USB_HOST_GEN_STATE_READY)
            {
                /* The instance is not ready for requests */
                genResult = USB_HOST_GEN_RESULT_BUSY;
            }
            else
            {
                if((size != 0) && (data == NULL))
                {
                    /* Input paramters are not valid */
                    genResult = USB_HOST_GEN_RESULT_INVALID_PARAMETER;
                }
                else
                {
                    /* The context for the transfer is the event that needs to
                     * be sent to the application. In this case the event to be
                     * sent to the application when the transfer completes is
                     * USB_HOST_GEN_EVENT_READ_COMPLETE */
                    
                    hostResult = USB_HOST_DeviceTransfer(genInstance->bulkInPipeHandle, tempTransferHandle, data, size, (uintptr_t)(USB_HOST_GEN_EVENT_READ_COMPLETE));
                    genResult = _USB_HOST_GEN_HostResutlToGENResultMap(hostResult);
                }
            }
        }
    }

    return(genResult);
}

// ****************************************************************************
/* Function:
    USB_HOST_GEN_RESULT USB_HOST_GEN_SerialStateNotificationGet
    (
        USB_HOST_GEN_HANDLE handle,
        USB_HOST_GEN_TRANSFER_HANDLE * transferHandle,
        USB_GEN_SERIAL_STATE * serialState
    );
           
  Summary:
    This function will request Serial State Notification from the attached
    device.

  Description:
    This function will request Serial State Notification from the attached
    device. If the request was accepted, transferHandle will contain a valid
    transfer handle, else it will contain USB_HOST_GEN_TRANSFER_HANDLE_INVALID.
    The completion of the request will be indicated by the
    USB_HOST_GEN_EVENT_SERIAL_STATE_NOTIFICATION_RECEIVED event. The transfer
    handle will be returned in the event. 

  Remarks:
    None.
*/

USB_HOST_GEN_RESULT USB_HOST_GEN_SerialStateNotificationGet
(
    USB_HOST_GEN_HANDLE handle,
    USB_HOST_GEN_TRANSFER_HANDLE * transferHandle,
    uint8_t * data
)
{
    USB_HOST_GEN_INSTANCE_OBJ * genInstance;
    USB_HOST_GEN_TRANSFER_HANDLE * tempTransferHandle, localTransferHandle;
    USB_HOST_GEN_RESULT genResult = USB_HOST_GEN_RESULT_FAILURE;
    USB_HOST_RESULT hostResult;

    genInstance = (USB_HOST_GEN_INSTANCE_OBJ *)handle;

    if(genInstance == NULL)
    {
        /* This handle is not valid */
        genResult = USB_HOST_GEN_RESULT_HANDLE_INVALID;
    }
    else
    {
        /* Check if the specified transfer handle holder is NULL, if so use a local
         * transfer handle holder */

        tempTransferHandle = (transferHandle == NULL) ? &localTransferHandle: transferHandle;

        if(!genInstance->inUse)
        {
            /* This object is not valid */
            genResult = USB_HOST_GEN_RESULT_DEVICE_UNKNOWN;
        }
        else
        {
            if(genInstance->state != USB_HOST_GEN_STATE_READY)
            {
                /* The instance is not ready for requests */
                genResult = USB_HOST_GEN_RESULT_BUSY;
            }
            else
            {
                if (data == NULL) 
                {
                    /* Input paramters are not valid */
                    genResult = USB_HOST_GEN_RESULT_INVALID_PARAMETER;
                }
                else
                {
                    /* The context for the transfer is the event that needs to
                     * be sent to the application. In this case the event to be
                     * sent to the application when the transfer completes is
                     * USB_HOST_GEN_EVENT_SERIAL_STATE_NOTIFICATION_RECEIVED */
                    
                    hostResult = USB_HOST_DeviceTransfer(genInstance->interruptPipeHandle, tempTransferHandle, data, 64, 
                            (uintptr_t)(USB_HOST_GEN_EVENT_SERIAL_STATE_NOTIFICATION_RECEIVED));
                    genResult = _USB_HOST_GEN_HostResutlToGENResultMap(hostResult);
                }
            }
        }
    }

    return(genResult);
}


